﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace Convolutional_Template
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
            ofd = new OpenFileDialog();
            sfd = new SaveFileDialog();
            ofd.Filter = filter;
            sfd.Filter = filter;
        }

        OpenFileDialog ofd;
        SaveFileDialog sfd;

        Bitmap image, processed, display;
        string filter = "Image | *.png; *.jpg; *.gif";
        string image_path;

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            sfd.FileName = image_path;
            if (sfd.ShowDialog() == DialogResult.OK)
            {
                processed.Save(sfd.FileName);
            }
        }

        private void filterToolStripMenuItem_Click(object sender, EventArgs e)
        {
            processed = processed.Filter(400);
            display = processed.ScaleImage(image_frame.Width, image_frame.Height);
            image_frame.Image = display;
        }

        private void loadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                image_path = ofd.FileName;
                image = (Bitmap)Bitmap.FromFile(image_path);
                processed = image;
                display = image.ScaleImage(image_frame.Width, image_frame.Height);
                image_frame.Image = display;
            }
        }
    }

    public static class Ext
    {
        public static Bitmap ScaleImage(this Bitmap image, int width, int height)
        {
            float xratio = (float)width / image.Width;
            float yratio = (float)height / image.Height;
            float ratio = Math.Min(xratio, yratio);

            int w = (int)(image.Width * ratio);
            int h = (int)(image.Height * ratio);

            Bitmap result = new Bitmap(image, new Size(w, h));
            return result;
        }

        public static Bitmap PadImage(this Bitmap image, int padding)
        {
            int w = image.Width;
            int h = image.Height;

            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.ReadOnly,
                PixelFormat.Format24bppRgb);

            int bytes = image_data.Stride * image_data.Height;
            byte[] buffer = new byte[bytes];

            Marshal.Copy(image_data.Scan0, buffer, 0, bytes);
            image.UnlockBits(image_data);

            int wpad = w + 2 * padding;
            int hpad = h + 2 * padding;

            Bitmap padded = new Bitmap(wpad, hpad);
            BitmapData padded_data = padded.LockBits(
                new Rectangle(0, 0, wpad, hpad),
                ImageLockMode.WriteOnly,
                PixelFormat.Format24bppRgb);

            int padded_bytes = padded_data.Stride * padded_data.Height;
            byte[] result = new byte[padded_bytes];
            for (int i = 0; i < padded_bytes; i++)
            {
                result[i] = 1;
            }

            for (int x = 0; x < w; x++)
            {
                for (int y = 0; y < h; y++)
                {
                    int img_pixel = x * 3 + y * image_data.Stride;
                    int pad_pixel = x * 3 + y * padded_data.Stride;
                    for (int c = 0; c < 3; c++)
                    {
                        result[padded_data.Stride * padding + 3 * padding + pad_pixel + c] = buffer[img_pixel + c];
                    }
                }
            }

            Marshal.Copy(result, 0, padded_data.Scan0, padded_bytes);
            padded.UnlockBits(padded_data);

            return padded;
        }

        public static double median(this byte[] array)
        {
            int len = array.Length;
            int mid_idx = len / 2;

            var sorted = array.OrderBy(n=>n);

            double median;

            if ((len % 2) == 0)
            {
                median = (sorted.ElementAt(mid_idx) + sorted.ElementAt(mid_idx - 1)) / 2;
            }

            else
            {
                median = sorted.ElementAt(mid_idx);
            }

            return median;
        }

        public static Bitmap Filter(this Bitmap image, double noise_variance)
        {
            int w = image.Width;
            int h = image.Height;

            BitmapData image_data = image.LockBits(
                new Rectangle(0, 0, w, h),
                ImageLockMode.ReadOnly,
                PixelFormat.Format24bppRgb);

            int bytes = image_data.Stride * image_data.Height;
            byte[] buffer = new byte[bytes];

            Marshal.Copy(image_data.Scan0, buffer, 0, bytes);
            image.UnlockBits(image_data);

            int r = 3;

            int wres = w - 2 * r;
            int hres = h - 2 * r;

            Bitmap result_image = new Bitmap(wres, hres);
            BitmapData result_data = result_image.LockBits(
                new Rectangle(0, 0, wres, hres),
                ImageLockMode.WriteOnly,
                PixelFormat.Format24bppRgb);
            int res_bytes = result_data.Stride * result_data.Height;
            byte[] result = new byte[res_bytes];

            for (int x = r; x < w - r; x++)
            {
                for (int y = r; y < h - r; y++)
                {

                    int pixel_location = x * 3 + y * image_data.Stride;
                    int res_pixel_loc = (x - r) * 3 + (y - r) * result_data.Stride;
                    double[] mean = new double[3];
                    int[][] neighborhood = new int[3][];

                    for (int c = 0; c < 3; c++)
                    {
                        neighborhood[c] = new int[(int)Math.Pow(2 * r + 1, 2)];
                        int added = 0;
                        for (int kx = -r; kx <= r; kx++)
                        {
                            for (int ky = -r; ky <= r; ky++)
                            {
                                int kernel_pixel = pixel_location + kx * 3 + ky * image_data.Stride;
                                neighborhood[c][added] = buffer[kernel_pixel + c];
                                added++;
                            }
                        }
                    }

                    for (int c = 0; c < 3; c++)
                    {
                        mean[c] = neighborhood[c].Sum() / Math.Pow(2 * r + 1, 2);
                        double variance = 0d;
                        for (int i = 0; i < neighborhood[c].Length; i++)
                        {
                            variance += Math.Pow(neighborhood[c][i] - mean[c], 2);
                        }
                        variance /= Math.Pow(2 * r + 1, 2);

                        if (noise_variance > variance)
                        {
                            noise_variance = variance;
                        }

                        result[res_pixel_loc + c] = (byte)(buffer[pixel_location + c] - (noise_variance / variance) * (buffer[pixel_location + c] - mean[c]));
                    }
                }
            }

            Marshal.Copy(result, 0, result_data.Scan0, res_bytes);
            result_image.UnlockBits(result_data);

            return result_image;
        }
    }
}
